/** 
 * File:   main.c
 * Author: Mathew Carr
 *
 * PSCD32 project.
 * 
 * Created on                                  28    August 2019, 22:40
 * 
 * Full CD32 joypad protocol support achieved   4 September 2019, 22:15
 * Mouse emulation mode programmed              7 September 2019, 01:17
 * Buttonjump mode programmed                  23 September 2019, 15:41
 * 
 * PS2 physical and logical interface and state symbols begin with PS2.
 * CD32 physical and logical interface and state symbols begin with CD32.
 * PSCD32 internal state symbols begin with PSCD32.
 */

/**
 * PSCD32 Software 
 * 
 * Copyright Mathew Carr 29th October 2019
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License
 * as listed in LICENSE.TXT
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
#include <xc.h>

/* Configuration bits to be written to reserved area. */
#pragma config POSCMOD  = NONE   /* Primary oscillator disabled. Use the internal FRCPLL. */
#pragma config IOL1WAY  = OFF    /* IOLOCK may be changed via unlocking seq. */
#pragma config OSCIOFNC = ON     /* OSC2/CLKO/RC15 functions as port I/O (RC15). */
#pragma config FCKSM    = CSECME /* Clock switching is enabled, Fail-Safe Clock Monitor is enabled. */
#pragma config FNOSC    = FRCPLL /* Internal Fast RC oscillator 8 MHz with 4x PLL = 32MHz. FCY = 16MIPS. */
#pragma config FWDTEN   = OFF    /* Watchdog Timer is disabled. */ /* TODO: THIS SHOULD BE ENABLED! */
#pragma config ICS      = PGx1   /* Emulator EMUC1/EMUD1 pins are shared with PGC1/PGD1. */
#pragma config GWRP     = OFF    /* Writes to program memory are enabled. */
#pragma config JTAGEN   = OFF    /* JTAG port is disabled. */

#define FCY  16000000L

#define HIGH 1
#define LOW  0

#define ABS(_x_) (((_x_) >= 0) ? (_x_) : (-(_x_)))

/** 
 * Macro for reversing the bit order of a byte.
 * Used to prepare messages towards PS2 since SPI always
 * sends MSB first.
 * Macro is used to prepare constant data during compilation!
 **/
#define INVERT_8BIT_ORDERING(_x_)   \
   (                                \
      (((_x_) & 0b00000001) << 7) | \
      (((_x_) & 0b00000010) << 5) | \
      (((_x_) & 0b00000100) << 3) | \
      (((_x_) & 0b00001000) << 1) | \
      (((_x_) & 0b00010000) >> 1) | \
      (((_x_) & 0b00100000) >> 3) | \
      (((_x_) & 0b01000000) >> 5) | \
      (((_x_) & 0b10000000) >> 7)   \
   )

/**
 * Array used to store array of bit-inverted bytes.
 * Used when receiving messages from PS2 since SPI
 * always receives MSB first.
 * Array is needed for look up of data in real-time.
 */
const unsigned char inverted_8bit_order_bytes[256] =
{
   0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 
   0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8, 
   0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 
   0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC, 
   0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2, 
   0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA, 
   0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6, 
   0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE, 
   0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1, 
   0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9, 
   0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5, 
   0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD, 
   0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 
   0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB, 
   0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7, 
   0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF, 
};

/**
 * Stores the complete state of a PS2 controller.
 * Buttons are stored as 1 idle, 0 pressed.
 * (Same 'switch grounded' set up as received from the pad.)
 */
typedef struct _PS2_CONTROLLER_STATE
{      
   union
   {
      struct {
         unsigned zero:8; /* Stores a constant 0 for CD32 shift reply. */
         unsigned one:1;  /* Stores a constant 1 for CD32 shift reply. */
   
         unsigned start:1;
         
         unsigned l1:1;
         unsigned r1:1;
   
         unsigned bS:1;
         unsigned bT:1;
         unsigned bX:1;
         unsigned bO:1;
      };
      unsigned int CD32reply; /* Contains a word 0bBRYGRLP100000000 */   
      
      /**
       * Construct the CD32 data word.
       * The SPI module outputs MSB first. 
       * The CD32 expects data in the following order:
       *  BLUE   (right button = O)         A CD32 strictly has a square shape:   GR  YE
       *  RED    (down button  = X)                                               RE  BL
       *  YELLOW (up button    = T)          but the RED button has the role of X on a PS2 pad.
       *  GREEN  (left button  = S)
       *  FF     (right shoulder)
       *  REW    (left shoulder)
       *  PAUSE  (center)
       *  1
       *  All future bits are zeroes.
       */
   };
   
   unsigned char u;
   unsigned char d;
   unsigned char l;
   unsigned char r;
   
   unsigned char select;
   
   unsigned char l2;
   unsigned char r2;
      
   unsigned char l3;
   unsigned char r3;
   
   unsigned char lx; /* 0x00 at upper left, 0xFF at lower right. */
   unsigned char ly;
   unsigned char rx;
   unsigned char ry;
   
   char peripheral_present; /* 0x00 for no controller, otherwise PS pad ID number from index 2 in response. */
   
   /* TODO: Individual button pressures? */
} PS2_CONTROLLER_STATE;

PS2_CONTROLLER_STATE global_controller_state;

#define PSCD32_OUTPUT_MODE_UNDEFINED                        0 /* On boot up, state undetermined. */
#define PSCD32_OUTPUT_MODE_CD32_JOYPAD_EMULATION            1 /* The dumb 2-button joystick / CD32 pad automatic switching JOYMODE mode. */
#define PSCD32_OUTPUT_MODE_AMIGA_TWO_BUTTON_JOYPAD_AUTOFIRE 2 /* Enhanced 2-button joystick with autofire. No CD32 support. */
#define PSCD32_OUTPUT_MODE_BUTTONJUMP_JOYPAD                3 /* D-Pad Up is disabled, X outputs UP. */
#define PSCD32_OUTPUT_MODE_MOUSE                            4 /* Analogue sticks control the mouse. */

typedef struct _PSCD32_state
{
   /* PSCD32_OUTPUT_MODE_* */
   int output_mode;
   
   /* Current quadrature frame. Need to maintain this state. */
   int mouse_x_quadrature_frame;
   int mouse_y_quadrature_frame;
   
   /* Timer 2 ticks before next quadrature frame advancement. */
   int mouse_x_controller_update_ticks_remaining;
   int mouse_y_controller_update_ticks_remaining;
   
   /* Timer 1 ticks used for timing autofires. */
   int fire1_autofire_timer;
   int fire2_autofire_timer;
} PSCD32_state;

PSCD32_state global_pscd32_state;

#define PERIPHERAL_TYPE_0x41_DIGITAL_CONTROLLER                        0x41
#define PERIPHERAL_TYPE_0x73_ANALOGUE_CONTROLLER                       0x73
#define PERIPHERAL_TYPE_0x23_NEGCON                                    0x23
#define PERIPHERAL_TYPE_0x53_GREEN_ANALOGUE_CONTROLLER_FLIGHTSTICK     0x53
#define PERIPHERAL_TYPE_0x12_MOUSE                                     0x12

/* TODO: This will need to be changed if I ever decide to detect button pressures. */
/* Length of SPI byte sequence STATUS for requesting data from PS1/2 pad. */
#define PS2_SPI_REQUEST_STATUS_LENGTH 9

/* Header response before pad type is determined. */
typedef struct __attribute__((packed)) _PERIPHERAL_RESPONSE_GENERIC
{
   unsigned char no_use;
   unsigned char peripheral_type;
   unsigned char spacer_0x5a;
} PERIPHERAL_RESPONSE_GENERIC;

/* Digital controller 0x41 response. */
typedef struct __attribute__((packed)) _PERIPHERAL_RESPONSE_0x41_DIGITAL_CONTROLLER
{
   PERIPHERAL_RESPONSE_GENERIC generic; 
   /* 1 16-bit word extra data after header. */
   union
   {
      struct __attribute__((packed))
      {
         unsigned select:1;
         unsigned blank2:1;
         unsigned blank1:1;
         unsigned start:1;
         unsigned u:1;
         unsigned r:1;
         unsigned d:1;
         unsigned l:1;
      };
      unsigned char dpad_buttons;            
   };
   union
   {
      struct __attribute__((packed))
      {
         unsigned l2:1;
         unsigned r2:1;
         unsigned l1:1;
         unsigned r1:1;
         unsigned bT:1;
         unsigned bO:1;
         unsigned bX:1;
         unsigned bS:1;
      };
      unsigned char shape_shoulder_buttons;
   };
} PERIPHERAL_RESPONSE_0x41_DIGITAL_CONTROLLER;

/* Analogue controller 0x73 response. */
typedef struct __attribute__((packed)) _PERIPHERAL_RESPONSE_0x73_ANALOGUE_CONTROLLER
{
   PERIPHERAL_RESPONSE_GENERIC generic; 
   /* 3 16-bit word extra data after header. */
   union
   {
      struct __attribute__((packed))
      {
         unsigned select:1;
         unsigned l3:1;
         unsigned r3:1;
         unsigned start:1;
         unsigned u:1;
         unsigned r:1;
         unsigned d:1;
         unsigned l:1;
      };
      unsigned char dpad_buttons;            
   };
   union
   {
      struct __attribute__((packed))
      {
         unsigned l2:1;
         unsigned r2:1;
         unsigned l1:1;
         unsigned r1:1;
         unsigned bT:1;
         unsigned bO:1;
         unsigned bX:1;
         unsigned bS:1;
      };
      unsigned char shape_shoulder_buttons;
   };
   unsigned char rx;
   unsigned char ry;
   unsigned char lx;
   unsigned char ly;
} PERIPHERAL_RESPONSE_0x73_ANALOGUE_CONTROLLER;

/* Bytes to send for STATUS. */
const unsigned char PS2_SPI_REQUEST_STATUS[PS2_SPI_REQUEST_STATUS_LENGTH] =
{
   INVERT_8BIT_ORDERING(0x01), INVERT_8BIT_ORDERING(0x42), /* Attention. Report your status. */
   0x00,       /* Peripherals return 0x5A here as a spacer. */
   0x00, 0x00, /* All peripherals report at least two bytes of data. */
   0x00, 0x00, /* Analogue controllers return two additional 16-bit words of data. */
   0x00, 0x00  
   /* TODO: Maybe longer messages to request button pressures? */
};

#define PS2_SPI_REQUEST_STATE_AL_BEFORE_DELAY     0 /* Do not send bytes, with attention low. */
#define PS2_SPI_REQUEST_STATE_TRANSFERRING_BYTES  1 /* Transferring a byte. */
#define PS2_SPI_REQUEST_STATE_AL_AFTER_DELAY      2 /* After sending message, with attention low. */
#define PS2_SPI_REQUEST_STATE_AH_DELAY            3 /* Delaying after transferring series of bytes, with attention high. */

/* Number of Timer1 resolution lengths to perform these following fixed-length states: */
#define PS2_SPI_REQUEST_AL_BEFORE_DELAY_LENGTH    2 /* Lower case D. */
#define PS2_SPI_REQUEST_AL_AFTER_DELAY_LENGTH     2 /* Lower case D. */
#define PS2_SPI_REQUEST_AH_DELAY_LENGTH           8 /* Upper case D. */

typedef union _PS2_SPI_PERIPHERAL_RESPONSE
{
   PERIPHERAL_RESPONSE_GENERIC                  header;
   PERIPHERAL_RESPONSE_0x41_DIGITAL_CONTROLLER  digital;
   PERIPHERAL_RESPONSE_0x73_ANALOGUE_CONTROLLER analogue;
   unsigned char raw_data[16]; /* TODO: Needs to change to accommodate button pressures if required. */
} PS2_SPI_PERIPHERAL_RESPONSE;

typedef struct _PS2_SPI_REQUEST
{
                           int  request_state;
   
   /* Number of events to repeat the current state action. */
                           int  events_remaining; 
                   
           const unsigned char *byte_tx_ptr;
                 unsigned char *byte_rx_ptr;
         
   PS2_SPI_PERIPHERAL_RESPONSE  received_data;
} PS2_SPI_REQUEST;

/* Global variable wrapping Timer1 and SPI1 for communication with PS2 pad. */
PS2_SPI_REQUEST ps2_spi_request;

/* Fill size bytes starting at dst with value. */
void memfill(unsigned char *dst, int size, unsigned char value)
{
   int i = 0;
   for (i = 0; i < size; i++)
   {
      *dst++ = value;
   }
}

/*
https://allpinouts.org/pinouts/connectors/input_device/sony-playstation-controller-port/  
  
The PSX Controller Data
Below are five tables that show the actual bytes sent by the controllers

Standard Digital Pad
        BYTE    CMND    DATA 
         01     0x01    idle 
         02     0x42    0x41 
         03     idle    0x5A    Bit0 Bit1 Bit2 Bit3 Bit4 Bit5 Bit6 Bit7 
         04     idle    data    SLCT           STRT UP   RGHT DOWN LEFT 
         05     idle    data    L2   R2    L1  R1   /\   O    X    |_| 
        All Buttons active low. 

NegCon
        BYTE    CMND    DATA 
         01     0x01    idle 
         02     0x42    0x23 
         03     idle    0x5A    Bit0 Bit1 Bit2 Bit3 Bit4 Bit5 Bit6 Bit7 
         04     idle    data                   STRT UP   RGHT DOWN LEFT 
         05     idle    data                   R1   A    B 
         06     idle    data    Steering  0x00 = Right 0xFF = Left 
         07     idle    data    I  Button 0x00 = Out   0xFF = In 
         08     idle    data    II Button 0x00 = Out   0xFF = In 
         09     idle    data    L1 Button 0x00 = Out   0xFF = In 
        All Buttons active low. 
  
Analogue Controller in Red Mode
        BYTE    CMND    DATA    * JOYL is definitely bit1, and JOYR is bit2.
         01     0x01    idle    * JOYL and JOYR are listed as WRONG in other docs!!
         02     0x42    0x73 
         03     idle    0x5A    Bit0 Bit1 Bit2 Bit3 Bit4 Bit5 Bit6 Bit7 
         04     idle    data    SLCT JOYL JOYR STRT UP   RGHT DOWN LEFT 
         05     idle    data    L2   R2   L1   R1   /\   O    X    |_| 
         06     idle    data    Right Joy 0x00 = Left  0xFF = Right 
         07     idle    data    Right Joy 0x00 = Up    0xFF = Down 
         08     idle    data    Left Joy  0x00 = Left  0xFF = Right 
         09     idle    data    Left Joy  0x00 = Up    0xFF = Down 
        All Buttons active low. 

Analogue Controller in Green Mode
        BYTE    CMND    DATA 
         01     0x01    idle 
         02     0x42    0x53 
         03     idle    0x5A    Bit0 Bit1 Bit2 Bit3 Bit4 Bit5 Bit6 Bit7 
         04     idle    data                   STRT UP   RGHT DOWN LEFT 
         05     idle    data    L2   L1   |_|  /\   R1   O    X    R2 
         06     idle    data    Right Joy 0x00 = Left  0xFF = Right 
         07     idle    data    Right Joy 0x00 = Up    0xFF = Down 
         08     idle    data    Left Joy  0x00 = Left  0xFF = Right 
         09     idle    data    Left Joy  0x00 = Up    0xFF = Down 
        All Buttons active low. 
  
PSX Mouse (credit to T.Fujita http://www.keisei.tsukuba.ac.jp/~kashima/games/ps-e.txt )
        BYTE    CMND    DATA 
         01     0x01    idle 
         02     0x42    0x12 
         03     idle    0x5A    Bit0 Bit1 Bit2 Bit3 Bit4 Bit5 Bit6 Bit7 
         04     idle    0xFF 
         05     idle    data         L    R 
         06     idle    data    Delta Vertical 
         07     idle    data    Delta Horizontal 
        All Buttons active low.  
 */

/**
 * Given a peripheral response byte sequence r, populate 
 * the memory controller state c with the extracted contents.
 * This also implicitly formats the face buttons into
 * the correct word format for the CD32 SPI shift.
 */
void PS2_CONTROLLER_STATE_analyse_received_data(PS2_CONTROLLER_STATE *c, const PS2_SPI_PERIPHERAL_RESPONSE *r)
{   
   if (r->header.spacer_0x5a != 0x5A)
   {
      /* No spacer means pad is not standard. */     
      memfill((unsigned char *)c, sizeof(PS2_CONTROLLER_STATE), 0x00);

      return;
   }
   
   c->peripheral_present = r->header.peripheral_type;
   
   /* These constants form part of the CD32 word. */
   c->one  = 1;
   c->zero = 0;
   
   switch (c->peripheral_present)
   {
      case (PERIPHERAL_TYPE_0x41_DIGITAL_CONTROLLER):
      {
         const PERIPHERAL_RESPONSE_0x41_DIGITAL_CONTROLLER *b = &(r->digital);
         
         c->u      = b->u;
         c->d      = b->d;
         c->l      = b->l;
         c->r      = b->r;

         c->select = b->select;
         c->start  = b->start;

         c->l3     = 1; /* Do not exist. */
         c->r3     = 1;
         
         c->l1     = b->l1;
         c->r1     = b->r1;
         c->l2     = b->l2;
         c->r2     = b->r2;

         c->bT     = b->bT;
         c->bO     = b->bO;
         c->bX     = b->bX;
         c->bS     = b->bS;

         c->lx     = 0x80; /* Do not exist. */
         c->ly     = 0x80;
         c->rx     = 0x80;
         c->ry     = 0x80;
      } break;
      case (PERIPHERAL_TYPE_0x73_ANALOGUE_CONTROLLER):
      {
         const PERIPHERAL_RESPONSE_0x73_ANALOGUE_CONTROLLER *b = &(r->analogue);
         
         c->u      = b->u;
         c->d      = b->d;
         c->l      = b->l;
         c->r      = b->r;

         c->select = b->select;
         c->start  = b->start;

         c->l3     = b->l3;
         c->r3     = b->r3;
         
         c->l1     = b->l1;
         c->r1     = b->r1;
         c->l2     = b->l2;
         c->r2     = b->r2;

         c->bT     = b->bT;
         c->bO     = b->bO;
         c->bX     = b->bX;
         c->bS     = b->bS;

         c->lx     = b->lx;
         c->ly     = b->ly;
         c->rx     = b->rx;
         c->ry     = b->ry;
      } break;
      default:
      {
         /* Unlucky, brother. Your pad is some strange one I don't know how to read yet. */
      } break;
   }   
}

void PS2_SPI_REQUEST_sequence_complete()
{         
   PS2_CONTROLLER_STATE_analyse_received_data(&global_controller_state, &(ps2_spi_request.received_data));
}

/**
 * Timer1 triggers every 8 + N SPI clock cycles. (N defined
 * by Timer period in timer set up.)
 * This is the resolution of the attention line/SPI request action.
 * This means that the ISR period is slightly longer than a SPI
 * transfer of a byte, so we can safely trigger a new byte after
 * the previous one has finished.
 * 
 * We use the ISR to trigger the byte transfers, as well as introduce
 * period-sized delays before and after sending the full message,
 * and a longer delay between messages with ATTENTION high.
 * 
 * The lower case ds are delays introduced while attention is low.
 * The upper case Ds are delays introduced between complete
 * communication exchanges to space them out. (I read somewhere
 * that the PS2 controller doesn't like being interrogated too
 * frequently.)
 * 
 * attention
 * ---___________---___________---
 * DDDd012345678dDDDd012345678dDDD
 * 
 * At any time, the contents of PS2_SPI_REQUEST contain
 * what should happen -AT THE NEXT INTERRUPT-.
 * 
 * Timer events
 * |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |
 * 
 * Get pad attention
 * --------____________________________________________--------____________
 * 
 * Trading data!
 *                
 * ------------Xvv-Xvv-Xvv-Xvv-Xvv-Xvv-Xvv-Xvv-Xvv-----------------Xvv-Xvv-
 *         sil.0   1   2   3   4   5   6   7   8   sil.        sil.0   1
 * 
 * SPI requests begin on Xs, and last for vv. The last
 * - symbol is idle time before the next Timer1 event happens.
 * There's a little attention time before and after the
 * complete message exchange for the pad to get ready.
 * sil. represents silence on the clock and data out lines.
 * Note that these messages are NINE BYTEs! The SPI system transfers bytes!
 */

#define PS2_SELECT_PIN_LATCH _LATB15

void __attribute__ ((interrupt, no_auto_psv))_T1Interrupt(void)
{
   /* Clear interrupt flag for Timer1 event. */
   IFS0bits.T1IF = 0;
   
   switch (ps2_spi_request.request_state)
   {
      case (PS2_SPI_REQUEST_STATE_AL_BEFORE_DELAY):
      {
         /* Pull attention down before transfers begin.*/
         PS2_SELECT_PIN_LATCH = 0;
         
         /* Prepare for the next transfer. */         
         /* TODO: Need some way to send things other than STATUS if I want to configure the pad?? */
         ps2_spi_request.byte_tx_ptr      = PS2_SPI_REQUEST_STATUS;
         ps2_spi_request.byte_rx_ptr      = ps2_spi_request.received_data.raw_data;
   
         ps2_spi_request.events_remaining--;
         
         if (ps2_spi_request.events_remaining == 0)
         {
            /* Next tick, we want to perform the transfers. */
            ps2_spi_request.request_state    = PS2_SPI_REQUEST_STATE_TRANSFERRING_BYTES;
            ps2_spi_request.events_remaining = PS2_SPI_REQUEST_STATUS_LENGTH;
         }
      } break;      
      case (PS2_SPI_REQUEST_STATE_TRANSFERRING_BYTES):
      {         
         /* Request to send the next byte. */
         SPI1BUF = *ps2_spi_request.byte_tx_ptr++;
         
         ps2_spi_request.events_remaining--;
         
         if (ps2_spi_request.events_remaining == 0)
         {            
            /* When all the bytes have been transferred, do an after delay. */
            ps2_spi_request.request_state    = PS2_SPI_REQUEST_STATE_AL_AFTER_DELAY;
            ps2_spi_request.events_remaining = PS2_SPI_REQUEST_AL_AFTER_DELAY_LENGTH;
         }
      } break;       
      case (PS2_SPI_REQUEST_STATE_AL_AFTER_DELAY):
      {
         ps2_spi_request.events_remaining--;
         
         if (ps2_spi_request.events_remaining == 0)
         {
            /* After this after-delay, go to the attention high state again. */
            ps2_spi_request.request_state    = PS2_SPI_REQUEST_STATE_AH_DELAY;
            ps2_spi_request.events_remaining = PS2_SPI_REQUEST_AH_DELAY_LENGTH;
            
            /**
             * This is a one-shot possibility that occurs immediately before
             * we make the pad idle. We have no need to communicate with the
             * pad in a timely manner any more because we have all the data:
             * it's a good time to analyse the retrieved state.
             */
            PS2_SPI_REQUEST_sequence_complete();
         }
      } break;   
      case (PS2_SPI_REQUEST_STATE_AH_DELAY):
      {
         PS2_SELECT_PIN_LATCH = 1;
         
         ps2_spi_request.events_remaining--;
         
         if (ps2_spi_request.events_remaining == 0)
         {
            /* When the AH state concludes on next tick, go low. */
            ps2_spi_request.request_state    = PS2_SPI_REQUEST_STATE_AL_BEFORE_DELAY;
            ps2_spi_request.events_remaining = PS2_SPI_REQUEST_AL_BEFORE_DELAY_LENGTH;
         }
      } break;      
      default:
      {
         /* TODO: Shouldn't happen? How to assert and report? */
      }
   }
}

void PS2_SPI_REQUEST_Timer1_SPI1_initialise_normal_request_loop(PS2_SPI_REQUEST *r)
{
   /**
    * Set the Timer1 ISR to act as if we've just
    * finished transferring a complete transaction.
    * We want it to linger high and begin a new transaction.
    */
   PS2_SELECT_PIN_LATCH = 1;
         
   ps2_spi_request.request_state    = PS2_SPI_REQUEST_STATE_AH_DELAY;
   ps2_spi_request.events_remaining = PS2_SPI_REQUEST_AH_DELAY_LENGTH;
}

/**
 * Fires right after a single byte exchange has taken place with PS2 pad.
 * Advance the RX pointer autonomously.
 * Don't actually send the next byte yet. The next Timer1 interrupt
 * handler will decide when and how to do that. This is how we
 * space out the transfers in time instead of having them back-to-back
 * immediately.
 */
void __attribute__ ((interrupt, no_auto_psv))_SPI1Interrupt(void)
{   
   *ps2_spi_request.byte_rx_ptr++ = inverted_8bit_order_bytes[SPI1BUF];
   
   /* Clear interrupt flag for SPI event. */
   IFS0bits.SPI1IF = 0;
}

/* GPIO DIP switches mounted on the board. */
#define DIP_SWITCH_1 (_RB14)
#define DIP_SWITCH_2 (_RB13)
#define DIP_SWITCH_3 (_RB12)
/* This is how they've ended up due to the way I've mounted them. */
#define DIP_SWITCH_LEFT  1 
#define DIP_SWITCH_RIGHT 0
/* Use these as boolean 1/0 values only, otherwise you'll end up comparing bitmasked values! */
#define DIP_SWITCH_1_IS_LEFT()   (DIP_SWITCH_1)
#define DIP_SWITCH_1_IS_RIGHT()  (!DIP_SWITCH_1_IS_LEFT())
#define DIP_SWITCH_2_IS_LEFT()   (DIP_SWITCH_2)
#define DIP_SWITCH_2_IS_RIGHT()  (!DIP_SWITCH_2_IS_LEFT())
#define DIP_SWITCH_3_IS_LEFT()   (DIP_SWITCH_3)
#define DIP_SWITCH_3_IS_RIGHT()  (!DIP_SWITCH_3_IS_LEFT())

#define CD32_SET_OUT_FIRE2OUT_DATAOUT_HIGH() { _LATB5  = 1; }
#define CD32_SET_OUT_FIRE2OUT_DATAOUT_LOW()  { _LATB5  = 0; }
#define CD32_SET_OUT_FIRE2OUT_DATAOUT(_x_)   { _LATB5  = (_x_); }

#define CD32_SET_OUT_UP_HIGH()               { _LATB6  = 1; }
#define CD32_SET_OUT_UP_LOW()                { _LATB6  = 0; }
#define CD32_SET_OUT_UP(_x_)                 { _LATB6  = (_x_); }

#define CD32_SET_OUT_DOWN_HIGH()             { _LATB7  = 1; }
#define CD32_SET_OUT_DOWN_LOW()              { _LATB7  = 0; }
#define CD32_SET_OUT_DOWN(_x_)               { _LATB7  = (_x_); }

#define CD32_SET_OUT_LEFT_HIGH()             { _LATB8  = 1; }
#define CD32_SET_OUT_LEFT_LOW()              { _LATB8  = 0; }
#define CD32_SET_OUT_LEFT(_x_)               { _LATB8  = (_x_); }

#define CD32_SET_OUT_RIGHT_HIGH()            { _LATB9  = 1; }
#define CD32_SET_OUT_RIGHT_LOW()             { _LATB9  = 0; }
#define CD32_SET_OUT_RIGHT(_x_)              { _LATB9  = (_x_); }

#define CD32_GET_IN_JOYMODE_FRAME()          (_RB10)

#define CD32_SET_OUT_FIREOUT_HIGH_Z()        { _TRISB11 = 1; }
#define CD32_SET_OUT_FIREOUT_OUTPUT()        { _TRISB11 = 0; }
#define CD32_SET_OUT_FIREOUT_HIGH()          { _LATB11 = 1; }
#define CD32_SET_OUT_FIREOUT_LOW()           { _LATB11 = 0; }
#define CD32_SET_OUT_FIREOUT(_x_)            { _LATB11 = (_x_); }
#define CD32_GET_IN_FIREOUT_CLOCKIN()        (_RB11)

void PSCD32_set_up_output_for_CD32_joypad_emulation();
void PSCD32_set_up_output_for_amiga_two_button_joypad_autofire();
void PSCD32_set_up_output_for_buttonjump_joypad();
void PSCD32_set_up_output_for_mouse();

void CD32_respond_as_CD32_joypad_emulation();
void CD32_respond_as_amiga_two_button_joypad_autofire();
void CD32_respond_as_buttonjump_joypad();
void CD32_respond_as_mouse();

/**
 * Timer2 fires at 1000Hz to update the UDLR Fire1 and Fire2 
 * outputs as if the pad were a dumb joystick.
 * 
 * We cannot manipulate Fire1 and Fire2 if the pad is currently
 * acting as a shift register due to the effect of JOYMODE.
 * 
 * To implement this, Timer2 is disabled during FRAME.
 * 
 * Or it fires at 16000Hz to update the UDLR as quadrature outputs
 * for the mouse, with Fire1 and Fire2 as the mouse buttons.
 */
void __attribute__ ((interrupt, no_auto_psv))_T2Interrupt(void)
{  
   /* Detect mode switch button combinations. */
   if (!global_controller_state.select)
   {
      if ((!global_controller_state.start)
       || (!global_controller_state.bX   ))
      {
         PSCD32_set_up_output_for_CD32_joypad_emulation();
      }
      else
      if (!global_controller_state.bO   )
      {
         PSCD32_set_up_output_for_amiga_two_button_joypad_autofire();
      }
      else
      if (!global_controller_state.bS   )
      {
         PSCD32_set_up_output_for_buttonjump_joypad();
      }
      else
      if (!global_controller_state.bT   )
      {
         PSCD32_set_up_output_for_mouse();
      }
   }
   else 
   {
      /* Do not update controller state if Select is held. */
      switch (global_pscd32_state.output_mode)
      {
         case (PSCD32_OUTPUT_MODE_CD32_JOYPAD_EMULATION):
         {
            CD32_respond_as_CD32_joypad_emulation();
         } break;
         case (PSCD32_OUTPUT_MODE_AMIGA_TWO_BUTTON_JOYPAD_AUTOFIRE):
         {
            CD32_respond_as_amiga_two_button_joypad_autofire();
         } break;      
         case (PSCD32_OUTPUT_MODE_BUTTONJUMP_JOYPAD):
         {
            CD32_respond_as_buttonjump_joypad();
         } break;
         case (PSCD32_OUTPUT_MODE_MOUSE):
         {
            CD32_respond_as_mouse();
         } break;
      }
   }
      
   /* Clear interrupt flag for Timer2 event. */
   IFS0bits.T2IF = 0;
}

/**
 * The CD32_SET_OUT_* construction will be compiled to a bit test
 * followed by a branch to branches containing a single bit test
 * or bit clear each.
 * If we attempted to set _RBx to a variable, a lot of masking
 * instructions would be generated.
 */

/* If a Fire is depressed, cancel autofire and just depress button. */
/* Otherwise if autofire is depressed, cycle through timers and depress button at 20Hz. (Full on/off cycle.) */
/* For 1000Hz, full cycle at 20Hz, we want 50 repeats per full cycle. */
   
#define TIMER_CYCLES_FOR_AUTOFIRE_CYCLE 50
#define TIMER_CYCLES_VALUE_AUTOFIRE_DEPRESSES_BUTTON_LIMIT (TIMER_CYCLES_FOR_AUTOFIRE_CYCLE/2)

/**
 * This is the size of the analogue stick active zone,
 * measured from the extremes of the stick,
 * to be read as a directional input from the player.
 * The stick is an unsigned value from 0x00 to 0xFF.
 * If this is 0x30, then < 0x30 is left, and > 0xCF is right.
 */
#define ANALOGUE_ACTIVE_ZONE_SIZE      0x30 
#define ANALOGUE_ACTIVE_AREA_U_L_THAN  (      ANALOGUE_ACTIVE_ZONE_SIZE)
#define ANALOGUE_ACTIVE_AREA_D_G_THAN  (0xFFu-ANALOGUE_ACTIVE_ZONE_SIZE)
#define ANALOGUE_ACTIVE_AREA_L_L_THAN  (      ANALOGUE_ACTIVE_ZONE_SIZE)
#define ANALOGUE_ACTIVE_AREA_R_G_THAN  (0xFFu-ANALOGUE_ACTIVE_ZONE_SIZE)

typedef struct __attribute__((packed)) _DIRECTIONAL_BITFIELDS
{
   unsigned char u;
   unsigned char d;
   unsigned char l;
   unsigned char r;
} DIRECTIONAL_BITFIELDS;

/**
 * Read the global_controller_state and combine the dpad and the
 * analogue stick button depressings into a single set of bitfields.
 * These are 1-idle and 0-pressed.
 */
DIRECTIONAL_BITFIELDS global_controller_state_get_dpad_analogue_stick_depressings()
{
   DIRECTIONAL_BITFIELDS bitfields;
   
   switch (global_controller_state.peripheral_present)
   {
      case (PERIPHERAL_TYPE_0x41_DIGITAL_CONTROLLER ) :
      case (PERIPHERAL_TYPE_0x73_ANALOGUE_CONTROLLER) :
      {   
         /* Treat digital pads as analogue as we force the phantom sticks centred. */
         bitfields.u = global_controller_state.u && (global_controller_state.ly > ANALOGUE_ACTIVE_AREA_L_L_THAN);
         bitfields.d = global_controller_state.d && (global_controller_state.ly < ANALOGUE_ACTIVE_AREA_R_G_THAN);
         bitfields.l = global_controller_state.l && (global_controller_state.lx > ANALOGUE_ACTIVE_AREA_L_L_THAN);
         bitfields.r = global_controller_state.r && (global_controller_state.lx < ANALOGUE_ACTIVE_AREA_R_G_THAN);
      } break;
      default:
      {
         bitfields.u = bitfields.d = bitfields.l = bitfields.r = 1;
      }
   }
   return bitfields;
}

void CD32_respond_as_CD32_joypad_emulation()
{        
   DIRECTIONAL_BITFIELDS directions = global_controller_state_get_dpad_analogue_stick_depressings();
   
   if (             directions.u ) { CD32_SET_OUT_UP_HIGH();               } else { CD32_SET_OUT_UP_LOW();               }
   if (             directions.d ) { CD32_SET_OUT_DOWN_HIGH();             } else { CD32_SET_OUT_DOWN_LOW();             }
   if (             directions.l ) { CD32_SET_OUT_LEFT_HIGH();             } else { CD32_SET_OUT_LEFT_LOW();             }
   if (             directions.r ) { CD32_SET_OUT_RIGHT_HIGH();            } else { CD32_SET_OUT_RIGHT_LOW();            }
      
   if (global_controller_state.bX) { CD32_SET_OUT_FIREOUT_HIGH();          } else { CD32_SET_OUT_FIREOUT_LOW();          }   
   if (global_controller_state.bO) { CD32_SET_OUT_FIRE2OUT_DATAOUT_HIGH(); } else { CD32_SET_OUT_FIRE2OUT_DATAOUT_LOW(); }   
}

void CD32_respond_as_amiga_two_button_joypad_autofire()
{   
   DIRECTIONAL_BITFIELDS directions = global_controller_state_get_dpad_analogue_stick_depressings();
   
   if (             directions.u ) { CD32_SET_OUT_UP_HIGH();               } else { CD32_SET_OUT_UP_LOW();               }
   if (             directions.d ) { CD32_SET_OUT_DOWN_HIGH();             } else { CD32_SET_OUT_DOWN_LOW();             }
   if (             directions.l ) { CD32_SET_OUT_LEFT_HIGH();             } else { CD32_SET_OUT_LEFT_LOW();             }
   if (             directions.r ) { CD32_SET_OUT_RIGHT_HIGH();            } else { CD32_SET_OUT_RIGHT_LOW();            }
   
   int fire1_depressed     = (!global_controller_state.bX || !global_controller_state.r1);
   int autofire1_depressed = (!global_controller_state.bS || !global_controller_state.l1);
   int emit_fire1          = 0;
   int fire2_depressed     = (!global_controller_state.bO || !global_controller_state.r2);
   int autofire2_depressed = (!global_controller_state.bT || !global_controller_state.l2);
   int emit_fire2          = 0;
   
   if (fire1_depressed)
   {
      global_pscd32_state.fire1_autofire_timer = 0;
      emit_fire1 = 1;
   }
   else
   if (autofire1_depressed)
   {
      if (++global_pscd32_state.fire1_autofire_timer >= TIMER_CYCLES_FOR_AUTOFIRE_CYCLE)
      {
         global_pscd32_state.fire1_autofire_timer = 0;
      }
      if (global_pscd32_state.fire1_autofire_timer < TIMER_CYCLES_VALUE_AUTOFIRE_DEPRESSES_BUTTON_LIMIT)
      {
         emit_fire1 = 1;
      }
   }
   else
   {
      global_pscd32_state.fire1_autofire_timer = 0;
   }
   
   if (fire2_depressed)
   {
      global_pscd32_state.fire2_autofire_timer = 0;
      emit_fire2 = 1;
   }
   else
   if (autofire2_depressed)
   {
      if (++global_pscd32_state.fire2_autofire_timer >= TIMER_CYCLES_FOR_AUTOFIRE_CYCLE)
      {
         global_pscd32_state.fire2_autofire_timer = 0;
      }
      if (global_pscd32_state.fire2_autofire_timer < TIMER_CYCLES_VALUE_AUTOFIRE_DEPRESSES_BUTTON_LIMIT)
      {
         emit_fire2 = 1;
      }
   }
   else
   {
      global_pscd32_state.fire2_autofire_timer = 0;
   }
      
   if (!emit_fire1) { CD32_SET_OUT_FIREOUT_HIGH();          } else { CD32_SET_OUT_FIREOUT_LOW();          }   
   if (!emit_fire2) { CD32_SET_OUT_FIRE2OUT_DATAOUT_HIGH(); } else { CD32_SET_OUT_FIRE2OUT_DATAOUT_LOW(); }   
}

void CD32_respond_as_buttonjump_joypad()
{     
   DIRECTIONAL_BITFIELDS directions = global_controller_state_get_dpad_analogue_stick_depressings();
      
   /* Analogue stick and d-pad no longer influence Up output. Down is special too. */   
   directions.u  = global_controller_state.bX;
   directions.u &= global_controller_state.r1;
   
   directions.d &= global_controller_state.r2;   
   
   if (             directions.u ) { CD32_SET_OUT_UP_HIGH();               } else { CD32_SET_OUT_UP_LOW();               }
   if (             directions.d ) { CD32_SET_OUT_DOWN_HIGH();             } else { CD32_SET_OUT_DOWN_LOW();             }
   if (             directions.l ) { CD32_SET_OUT_LEFT_HIGH();             } else { CD32_SET_OUT_LEFT_LOW();             }
   if (             directions.r ) { CD32_SET_OUT_RIGHT_HIGH();            } else { CD32_SET_OUT_RIGHT_LOW();            }
        
   int fire1_depressed     = (!global_controller_state.bO);
   int autofire1_depressed = (!global_controller_state.bS);
   int emit_fire1          = 0;
   int emit_fire2          = (!global_controller_state.bT);
   
   if (fire1_depressed)
   {
      global_pscd32_state.fire1_autofire_timer = 0;
      emit_fire1 = 1;
   }
   else
   if (autofire1_depressed)
   {
      if (++global_pscd32_state.fire1_autofire_timer >= TIMER_CYCLES_FOR_AUTOFIRE_CYCLE)
      {
         global_pscd32_state.fire1_autofire_timer = 0;
      }
      if (global_pscd32_state.fire1_autofire_timer < TIMER_CYCLES_VALUE_AUTOFIRE_DEPRESSES_BUTTON_LIMIT)
      {
         emit_fire1 = 1;
      }
   }
   else
   {
      global_pscd32_state.fire1_autofire_timer = 0;
   }
   
   if (!emit_fire1) { CD32_SET_OUT_FIREOUT_HIGH();          } else { CD32_SET_OUT_FIREOUT_LOW();          }   
   if (!emit_fire2) { CD32_SET_OUT_FIRE2OUT_DATAOUT_HIGH(); } else { CD32_SET_OUT_FIRE2OUT_DATAOUT_LOW(); }   
}

void CD32_respond_as_mouse()
{
   /* When dpad is depressed, use this as the analogue stick value. */
#define ANALGOUE_MOUSE_DPAD_SIMULATED_VALUE 0x07   
   
   int analogue_lx = global_controller_state.lx;
   
   if (!global_controller_state.l) analogue_lx =        ANALGOUE_MOUSE_DPAD_SIMULATED_VALUE;
   if (!global_controller_state.r) analogue_lx = 0xFF - ANALGOUE_MOUSE_DPAD_SIMULATED_VALUE;
   
   int analogue_ly = global_controller_state.ly;
   
   if (!global_controller_state.u) analogue_ly =        ANALGOUE_MOUSE_DPAD_SIMULATED_VALUE;
   if (!global_controller_state.d) analogue_ly = 0xFF - ANALGOUE_MOUSE_DPAD_SIMULATED_VALUE;
   
#define ANALOGUE_MOUSE_DEAD_ZONE_SIZE 0x20
#define ANALOGUE_MOUSE_DEAD_ZONE_EXTENT_R (0x80+ANALOGUE_MOUSE_DEAD_ZONE_SIZE)
#define ANALOGUE_MOUSE_DEAD_ZONE_EXTENT_L (0x80-ANALOGUE_MOUSE_DEAD_ZONE_SIZE)
   
   if ((analogue_lx < ANALOGUE_MOUSE_DEAD_ZONE_EXTENT_R)   
    && (analogue_lx > ANALOGUE_MOUSE_DEAD_ZONE_EXTENT_L))
   {
      /* Mouse X within dead zone. Stop mouse X. Don't advance timers or quad frames. */
      global_pscd32_state.mouse_x_controller_update_ticks_remaining = 0;
   }
   else
   {   
      int analogue_x_quad_direction = (analogue_lx < 0x80) ? -1 : +1;
   
      if (global_pscd32_state.mouse_x_controller_update_ticks_remaining == 0)
      {
         global_pscd32_state.mouse_x_quadrature_frame += analogue_x_quad_direction;
         global_pscd32_state.mouse_x_quadrature_frame &= 3;
         
         switch (global_pscd32_state.mouse_x_quadrature_frame)
         {
            case (0): { CD32_SET_OUT_DOWN_HIGH(); CD32_SET_OUT_RIGHT_HIGH(); } break;
            case (1): { CD32_SET_OUT_DOWN_LOW();  CD32_SET_OUT_RIGHT_HIGH(); } break;
            case (2): { CD32_SET_OUT_DOWN_LOW();  CD32_SET_OUT_RIGHT_LOW();  } break;
            case (3): { CD32_SET_OUT_DOWN_HIGH(); CD32_SET_OUT_RIGHT_LOW();  } break;
         }
         
         int analogue_deflection_x = ABS(analogue_lx - 0x80);
         
         global_pscd32_state.mouse_x_controller_update_ticks_remaining = 1 * (0x88 - analogue_deflection_x);
      }
      else
      {
         global_pscd32_state.mouse_x_controller_update_ticks_remaining--;
      }
   }
   
   if ((analogue_ly < ANALOGUE_MOUSE_DEAD_ZONE_EXTENT_R)   
    && (analogue_ly > ANALOGUE_MOUSE_DEAD_ZONE_EXTENT_L))
   {
      /* Mouse Y within dead zone. Stop mouse Y. Don't advance timers or quad frames. */
      global_pscd32_state.mouse_y_controller_update_ticks_remaining = 0;
   }
   else
   {   
      int analogue_y_quad_direction = (analogue_ly < 0x80) ? -1 : +1;
   
      if (global_pscd32_state.mouse_y_controller_update_ticks_remaining == 0)
      {
         global_pscd32_state.mouse_y_quadrature_frame += analogue_y_quad_direction;
         global_pscd32_state.mouse_y_quadrature_frame &= 3;
         
         switch (global_pscd32_state.mouse_y_quadrature_frame)
         {
            case (0): { CD32_SET_OUT_UP_HIGH(); CD32_SET_OUT_LEFT_HIGH(); } break;
            case (1): { CD32_SET_OUT_UP_LOW();  CD32_SET_OUT_LEFT_HIGH(); } break;
            case (2): { CD32_SET_OUT_UP_LOW();  CD32_SET_OUT_LEFT_LOW();  } break;
            case (3): { CD32_SET_OUT_UP_HIGH(); CD32_SET_OUT_LEFT_LOW();  } break;
         }
         
         int analogue_deflection_y = ABS(analogue_ly - 0x80);
         
         global_pscd32_state.mouse_y_controller_update_ticks_remaining = 1 * (0x88 - analogue_deflection_y);
      }
      else
      {
         global_pscd32_state.mouse_y_controller_update_ticks_remaining--;
      }
   }
   
   if (global_controller_state.bX && global_controller_state.l1) { CD32_SET_OUT_FIREOUT_HIGH();          } else { CD32_SET_OUT_FIREOUT_LOW();          }
   if (global_controller_state.bO && global_controller_state.r1) { CD32_SET_OUT_FIRE2OUT_DATAOUT_HIGH(); } else { CD32_SET_OUT_FIRE2OUT_DATAOUT_LOW(); }
}

/**
 * External Interrupt 1 is a falling edge on JOYMODE/FRAME.
 * The CD32 is interrogating the shift register.
 * FIRE1OUT/CLOCKIN becomes HIGH-Z input for the clock.
 * We want to enable the SPI2 to take control of the appropriate
 * output pins.
 * This is only active if the PSCD32 is in joystick mode, not mouse.
 */
void __attribute__ ((interrupt, no_auto_psv))_INT1Interrupt(void)
{      
   /* Disable Timer2 to prevent farting over our nice bidirectional states. */
   T2CONbits.TON        = 0;
   /* Set bidirectional clock pin to input. */
   CD32_SET_OUT_FIREOUT_HIGH_Z();
   
   /* Enable SPI for new transmission. */
   SPI2STATbits.SPIEN   = 1;
   
   /* Put in my word of data. See the comment in PS2_CONTROLLER_STATE for how this works. */
   SPI2BUF = global_controller_state.CD32reply;
      
   /* Clear overflow and SPI2 event flags in case that jams things up. */
   IFS2bits.SPI2IF      = 0;
   SPI2STATbits.SPIROV  = 0;     
   
   /* Clear interrupt flag. */
   IFS1bits.INT1IF      = 0;
}   

/**
 * External Interrupt 2 is a rising edge on JOYMODE/FRAME.
 * The CD32 is returning to normal dumb joystick mode.
 */
void __attribute__ ((interrupt, no_auto_psv))_INT2Interrupt(void)
{
   /* Disable SPI, and clear flag. */
   SPI2STATbits.SPIEN   = 0;
   IFS1bits.INT2IF      = 0;
   
   /* Set bidirectional pin to output. */
   CD32_SET_OUT_FIREOUT_OUTPUT();
   
   /* Re-enable Timer2 to allow updating ULDR. */
   T2CONbits.TON        = 1;
}

/**
 * Fires right after a single word exchange has taken place with Amiga system.
 * Disable SPI because we don't have any more data to send.
 * Disabling the modules sets the output CLOCKEDDATAOUT line to low.
 * This would only be fired if the Amiga clocks a CD32 pad for
 * sixteen cycles for an unknown reason.
 */
void __attribute__ ((interrupt, no_auto_psv))_SPI2Interrupt(void)
{
   /* Disable SPI, and clear flag. */
   SPI2STATbits.SPIEN   = 0;      
   IFS2bits.SPI2IF      = 0;
}

void PSCD32_set_up_output_for_CD32_joypad_emulation()
{
   if (global_pscd32_state.output_mode == PSCD32_OUTPUT_MODE_CD32_JOYPAD_EMULATION) return;
   
   global_pscd32_state.fire1_autofire_timer = 0;
   global_pscd32_state.fire2_autofire_timer = 0;
   
   /* Enable the interrupts that allow the CD32 to force the PSCD32 into shift register mode. */
   IFS1bits.INT1IF      = 0;     /* Clear flags. */
   IFS1bits.INT2IF      = 0;
   IEC1bits.INT1IE      = 1;     /* Enable interrupts. */
   IEC1bits.INT2IE      = 1;

   IFS2bits.SPI2IF      = 0;     /* Clear interrupt flag for SPI event. */
   IEC2bits.SPI2IE      = 1;     /* Enable interrupt request from SPI. */
   
   /* Timer operates at slow rate for just joystick input. 
    * Joystick =  1,000Hz. Ratio = 16,000,000/ 1,000 = 16,000 period value.
    * Mouse    = 10,000Hz. Ratio = 16,000,000/10,000 =  1,600 period value.
    */   
   PR2                  = 16000-1; /* Period value for Timer2. This value is reached, proceeds, and then TMR2 is set to 0. */
   TMR2                 = 0;       /* Reset timer value. */
      
   global_pscd32_state.output_mode = PSCD32_OUTPUT_MODE_CD32_JOYPAD_EMULATION;
}

void PSCD32_set_up_output_for_amiga_two_button_joypad_autofire()
{
   if (global_pscd32_state.output_mode == PSCD32_OUTPUT_MODE_AMIGA_TWO_BUTTON_JOYPAD_AUTOFIRE) return;
      
   global_pscd32_state.fire1_autofire_timer = 0;
   global_pscd32_state.fire2_autofire_timer = 0;
   
   /* Disable the interrupts that allow the CD32 to force the PSCD32 into shift register mode. */
   IFS1bits.INT1IF      = 0;     /* Clear flags. */
   IFS1bits.INT2IF      = 0;
   IEC1bits.INT1IE      = 0;     /* Disable interrupts. */
   IEC1bits.INT2IE      = 0;

   IFS2bits.SPI2IF      = 0;     /* Clear interrupt flag for SPI event. */
   IEC2bits.SPI2IE      = 0;     /* Disable interrupt request from SPI. */
   
   /* Timer operates at slow rate for just joystick input. 
    * Joystick =  1,000Hz. Ratio = 16,000,000/ 1,000 = 16,000 period value.
    * Mouse    = 10,000Hz. Ratio = 16,000,000/10,000 =  1,600 period value.
    */   
   PR2                  = 16000-1; /* Period value for Timer2. This value is reached, proceeds, and then TMR2 is set to 0. */
   TMR2                 = 0;       /* Reset timer value. */
      
   global_pscd32_state.output_mode = PSCD32_OUTPUT_MODE_AMIGA_TWO_BUTTON_JOYPAD_AUTOFIRE;
}

void PSCD32_set_up_output_for_buttonjump_joypad()
{
   if (global_pscd32_state.output_mode == PSCD32_OUTPUT_MODE_BUTTONJUMP_JOYPAD) return;
      
   global_pscd32_state.fire1_autofire_timer = 0;
   global_pscd32_state.fire2_autofire_timer = 0;
   
   /* Disable the interrupts that allow the CD32 to force the PSCD32 into shift register mode. */
   IFS1bits.INT1IF      = 0;     /* Clear flags. */
   IFS1bits.INT2IF      = 0;
   IEC1bits.INT1IE      = 0;     /* Disable interrupts. */
   IEC1bits.INT2IE      = 0;

   IFS2bits.SPI2IF      = 0;     /* Clear interrupt flag for SPI event. */
   IEC2bits.SPI2IE      = 0;     /* Disable interrupt request from SPI. */
   
   /* Timer operates at slow rate for just joystick input. 
    * Joystick =  1,000Hz. Ratio = 16,000,000/ 1,000 = 16,000 period value.
    * Mouse    = 10,000Hz. Ratio = 16,000,000/10,000 =  1,600 period value.
    */   
   PR2                  = 16000-1; /* Period value for Timer2. This value is reached, proceeds, and then TMR2 is set to 0. */
   TMR2                 = 0;       /* Reset timer value. */
      
   global_pscd32_state.output_mode = PSCD32_OUTPUT_MODE_BUTTONJUMP_JOYPAD;
}

void PSCD32_set_up_output_for_mouse()
{
   if (global_pscd32_state.output_mode == PSCD32_OUTPUT_MODE_MOUSE) return;
      
   global_pscd32_state.mouse_x_quadrature_frame = 0;
   global_pscd32_state.mouse_y_quadrature_frame = 0;
   
   global_pscd32_state.mouse_x_controller_update_ticks_remaining = 0;
   global_pscd32_state.mouse_y_controller_update_ticks_remaining = 0;
   
   /* Disable the interrupts that allow the CD32 to force the PSCD32 into shift register mode. */
   IFS1bits.INT1IF      = 0;     /* Clear flags. */
   IFS1bits.INT2IF      = 0;
   IEC1bits.INT1IE      = 0;     /* Disable interrupts. */
   IEC1bits.INT2IE      = 0;

   IFS2bits.SPI2IF      = 0;     /* Clear interrupt flag for SPI event. */
   IEC2bits.SPI2IE      = 0;     /* Disable interrupt request from SPI. */
   
   /* Timer operates at fast rate for quadrature output.
    * Joystick =  1,000Hz. Ratio = 16,000,000/ 1,000 = 16,000 period value.
    * Mouse    = 10,000Hz. Ratio = 16,000,000/10,000 =  1,600 period value.
    */   
   PR2                  = 1600-1;  /* Period value for Timer2. This value is reached, proceeds, and then TMR2 is set to 0. */
   TMR2                 = 0;       /* Reset timer value. */
   
   /* TODO: Set JOYMODE/FRAME to be an output pin if we want to be gratuitous and support middle mouse. */
   
   global_pscd32_state.output_mode = PSCD32_OUTPUT_MODE_MOUSE;
}

/*
 * Entry point after reset.
 */
int main(int argc, char **argv)
{
   /* Do nothing until the PLL lock has been achieved. */
   while (!OSCCONbits.LOCK) {};
   
   /* Clear all the reset status bits.
      This also contains the SWDTEN software watchdog control, and
      the Program Memory Power During Sleep bit.
      Sections 9.2 and 24.2.5 say I should leave this set for
      faster waking.
    */
   RCON &= 0b0000000100000000;
         
   /* Set the state machine to begin as if we're just entering the AH high phase. */
   PS2_SPI_REQUEST_Timer1_SPI1_initialise_normal_request_loop(&ps2_spi_request);
   
   /* Power down all the peripherals, except SPI1, SPI2 and Timers */
   PMD1 = 0b0000000011100001; /* Enable timers, SPI1, SPI2.  Disable I2C module 1, U2, U1, ADC1. U1 and U2 are UARTs. */
   PMD2 = 0b0001111100011111; /* Enable nothing. Disable Input Capture and Output Compare. */
   PMD3 = 0b0000011110000010; /* Enable nothing. Disable Comparator module, Real-Time Clock, Parallel Master Port, CRC check, I2C module 2. */
      
   /* Safely initialise the chip. We want everything to be inputs first! */
   /* Pin assignments:
      Pin  1: !MCLR
      Pin  2: N/C                                     RA0 unused outputting low
      Pin  3: N/C                                     RA1 unused outputting low
      Pin  4: PICkit PGED1
      Pin  5: PICkit PGEC1
      Pin  6: PS2 SPI OUT COMMAND MOSI                RP2 assigned to SPI1 MOSI
      Pin  7: PS2 SPI OUT CLOCK                       RP3 assigned to SPI1 CLOCK
      Pin  8: VSS
      Pin  9: HI-Z                                    RA2 as input
      Pin 10: PS2 ACKNOWLEDGE IN momentary low pulse  RA3 as input, possibly use CN29 interrupt source
      Pin 11: PS2 IN DATA MISO                        RB4 digital in assigned to MISO
      Pin 12: N/C                                     RA4 unused outputting low
      Pin 13: VDD
      Pin 14: CD32 FIRE2OUT/DATA OUT                  RB5 as output, open drain
    */      
   /* !MCLR */
   _TRISA0  = 0; _LATA0 = 0;             /* Unused, not connected. */
   _TRISA1  = 0; _LATA1 = 0;             /* Unused, not connected. */   
   _TRISB0  = 0; _LATB0 = 0;             /* PICkit dedicated, otherwise not connected. */
   _TRISB1  = 0; _LATB1 = 0;             /* PICkit dedicated, otherwise not connected. */
   _TRISB2  = 0; _LATB2 = 1;             /* PS2 SPI MOSI Idle high output digital signal. */
   _TRISB3  = 0; _LATB3 = 1;             /* PS2 SPI CLOCK Idle high output digital signal. */
   /* VSS */
   _TRISA2  = 1;                         /* Unused - trace flows through, keep it input. */
   _TRISA3  = 1;                         /* PS2 acknowledge input. */
   _TRISB4  = 1;                         /* PS2 SPI MOSI input. */
   _TRISA4  = 0; _LATA4 = 0;             /* Unused, not connected. */
   /* VDD */
   _TRISB5  = 0; _LATB5 = 0; _ODB5 = 0;  /* CD32 Fire2out / Data out open drain. */
      
   /* Pin 28: VDD
      Pin 27: VSS
      Pin 26: PS2 SPI OUT SSELECT                     RB15 manually GPIO controlled to act as SPI1 SSELECT
      Pin 25: INTERNAL PU DIP OPTION 1 TO GROUND      RB14 as input, PU set
      Pin 24: INTERNAL PU DIP OPTION 2 TO GROUND      RB13 as input, PU set
      Pin 23: INTERNAL PU DIP OPTION 3 TO GROUND      RB12 as input, PU set
      Pin 22: CD32 FIREOUT/CLOCK IN                   Bidirectional: either open drain output for Fire as RB9 GPIO, or slave select hiZ input for RP9 SCK2
      Pin 21: CD32 JOYMODE IN                         RB10 as input, external interrupt
      Pin 20: VCAP
      Pin 19: VSS
      Pin 18: CD32 RIGHT OUT                          RB9 as output, open drain
      Pin 17: CD32 LEFT OUT                           RB8 as output, open drain
      Pin 16: CD32 DOWN OUT                           RB7 as output, open drain
      Pin 15: CD32 UP OUT                             RB6 as output, open drain
   */
   /* Set pins 14,15,16,17,18 to be open drain output. */
   /* Pin 22 changes between open drain and input. */
   
   /* VDD */
   /* VSS */
   _TRISB15 = 0; _LATB15 = 1;             /* PS2 GPIO controlled ATTENTION Idle high output digital signal. */
   _TRISB14 = 1; _CN12PUE = 1;            /* DIP input internally pulled up switch to ground. */
   _TRISB13 = 1; _CN13PUE = 1;            /* DIP input internally pulled up switch to ground. */
   _TRISB12 = 1; _CN14PUE = 1;            /* DIP input internally pulled up switch to ground. */
   _TRISB11 = 0; _LATB11 = 0; _ODB11 = 0; /* CD32 bidirectional..!! */
   _TRISB10 = 1;                          /* CD32 Joymode input. */
   /* VCAP */
   /* VSS */   
   _TRISB9  = 0; _LATB9 = 0; _ODB9 = 1;   /* CD32 Right out open drain. */
   _TRISB8  = 0; _LATB8 = 0; _ODB8 = 1;   /* CD32 Left out open drain. */
   _TRISB7  = 0; _LATB7 = 0; _ODB7 = 1;   /* CD32 Down out open drain. */
   _TRISB6  = 0; _LATB6 = 0; _ODB6 = 1;   /* CD32 Up out open drain. */
   
   CD32_SET_OUT_UP_HIGH();
   CD32_SET_OUT_DOWN_HIGH();
   CD32_SET_OUT_LEFT_HIGH();
   CD32_SET_OUT_RIGHT_HIGH();
   CD32_SET_OUT_FIREOUT_HIGH();
   CD32_SET_OUT_FIREOUT_OUTPUT();
   CD32_SET_OUT_FIRE2OUT_DATAOUT_HIGH();
      
   if (1) /* Set up Timer1 and SPI1 together PS2 communication. */
   {
      /**
       * Timer1 triggers at coarse rate to delegate SPI byte sendings.
       * 
       * SPI clock set as 166.667 kHz.
       * We want Timer1 actions to be produced every 8+12 SPI clock cycles.
       * 166.667kHz / 20 cycles = 8333.3Hz
       * 
       * FCY = 16 MIPS. 
       * Ratio = 16,000,000/12,820.5 = 1,920
       */
      /* Disable interrupts for safety, and clear interrupt flag. */
      IFS0bits.T1IF        = 0;    
      IEC0bits.T1IE        = 0;
      
      T1CONbits.TON        = 0;        /* Timer off. */
      T1CONbits.TSIDL      = 0;        /* Continue in idle. */
      T1CONbits.TGATE      = 0;        /* Gated time accumulation disabled. (What even is it?) */
      T1CONbits.TCKPS      = 0b00;     /* 1:1 prescale. */
      T1CONbits.TSYNC      = 0;        /* Don't use external clock, set this to zero. */
      T1CONbits.TCS        = 0;        /* Internal clock source FOSC/2. */
      PR1                  = (1920-1); /* Period value for Timer1. This value is reached, proceeds, and then TMR1 is set to 0. */
      TMR1                 = 0;        /* Reset timer value. */
   
      /* SPI setup for PS2 communications on left side. */      
      /* Disable SPI module before configuring. */
      SPI1STATbits.SPIEN   = 0;  
      
      /* Set the TRIS of the SPI pins to the correct state. */
      _TRISB2  = 0; /* MOSI    */
      _TRISB3  = 0; /* CLOCK   */
      _TRISB4  = 1; /* MISO    */
      _TRISB15 = 0; /* SSELECT */
      
      /* Bit of OSCCON register controlling peripheral remapping lock. */
      const unsigned char OSCCON_LOCK = 0b01000000;
      
      /* Unlock the peripheral remapping. */
      __builtin_write_OSCCONL(OSCCON & ~(OSCCON_LOCK));
      
      /* Assign the remappable peripheral pins OUTPUT registers according to my layout: */
      _RP2R    = 7; /* MOSI    output to RP2 pin. */
      _RP3R    = 8; /* CLOCK   output to RP3 pin. */
      _RP4R    = 0; /* MISO    RP4  pin is used for MISO input, don't direct any output there. */
      _RP15R   = 0; /* SSELECT RP15 pin is used for ATTENTION GPIO output, don't direct any SPI-module output there. */      
      /* Assign the remappable peripheral pins INPUT, RP4 is input for SPI1 slave data. */
      _SDI1R   = 4; /* SPI1 MISO input is directed to read from RP4. */

      /* Lock the peripheral remapping. */
      __builtin_write_OSCCONL(OSCCON | OSCCON_LOCK);

      /* Disable interrupts for safety, and clear interrupt flag. */
      IFS0bits.SPI1IF      = 0;
      IEC0bits.SPI1IE      = 0;
      
      /* These are zero by default anyway... */
      SPI1STATbits.SPISIDL = 0;     /* Continue in idle mode. */
      SPI1STATbits.SPIROV  = 0;     /* Clear overflow flag.   */
            
      SPI1CON1bits.DISSCK  = 0;     /* Enable clock output. */
      SPI1CON1bits.DISSDO  = 0;     /* Enable data output. */      
      SPI1CON1bits.MODE16  = 0;     /* Exchange 8-bit words. */
      SPI1CON1bits.SMP     = 0;     /* Sample in middle of clock cycle. */
      SPI1CON1bits.CKE     = 0;     /* Output data is changed when clock cycle begins. */
      SPI1CON1bits.SSEN    = 0;     /* SSx pin is not used. We manually control ATTENTION as GPIO low over multiple byte transfers. */
      SPI1CON1bits.CKP     = 1;     /* Clock is high idle, low active. */
      SPI1CON1bits.MSTEN   = 1;     /* PIC24 chip is SPI Master. */
      SPI1CON2bits.FRMEN   = 0;     /* Framed mode disabled. It isn't used in this protocol and not supported by the PIC24 anyway. */
      
      /**
       * We want a 150kHz clock rate to the PS2 pad.
       * FCY  = FOSC / 2 = 32,000,000 Hz / 2 = 16,000,000 Hz.
       * FSCK = 150kHz                       =    150,000 Hz.
       * Quotient = 106.667.
       * Product of prescalers is 96 = 16 * 6.
       * Effective clock rate: 16,000,000/96 =    166,666 Hz.
       */
      SPI1CON1bits.SPRE    = 0b010; /* Secondary prescaler is  6. */
      SPI1CON1bits.PPRE    =  0b01; /*   Primary prescaler is 16. */      
         
      /* We're done so we can turn everything on now. */
      IFS0bits.SPI1IF      = 0;     /* Clear interrupt flag for SPI event. */
      IEC0bits.SPI1IE      = 1;     /* Enable interrupt request from SPI. */
      SPI1STATbits.SPIEN   = 1;     /* Enable SPI module. */      
      
      IFS0bits.T1IF        = 0;     /* Clear interrupt flag for Timer1 event. */   
      IEC0bits.T1IE        = 1;     /* Enable interrupt for Timer1. */  
      T1CONbits.TON        = 1;     /* Timer on. */
   }    
         
   if (1) /* Set up Timer2. */
   {
      /**
       * Timer2 triggers to update UDLR and Fire1 and 2 in
       * dumb mode. It also needs to advance the quadrature timers in
       * mouse mode.
       * 
       * Joystick =  1,000Hz. Ratio = 16,000,000/ 1,000 = 16,000 period value.
       * Mouse    = 10,000Hz. Ratio = 16,000,000/10,000 =  1,600 period value.
       * 
       */
      /* Disable interrupts for safety, and clear interrupt flag. */
      IFS0bits.T2IF        = 0;    
      IEC0bits.T2IE        = 0;
      
      T2CONbits.TON        = 0;         /* Timer off. */
      T2CONbits.TSIDL      = 0;         /* Continue in idle. */
      T2CONbits.TGATE      = 0;         /* Gated time accumulation disabled. (What even is it?) */
      T2CONbits.TCKPS      = 0b00;      /* 1:1 prescale. */
      T2CONbits.T32        = 0;         /* Don't combine timers. */
      T2CONbits.TCS        = 0;         /* Internal clock source FOSC/2. */
      PR2                  = 65534;     /* Period value for Timer2. This value is reached, proceeds, and then TMR2 is set to 0. */
      TMR2                 = 0;         /* Reset timer value. */
      /* Timer period is set by set up output function. */
      
      IEC0bits.T2IE        = 1;         /* Enable interrupt for Timer2. */  
      T2CONbits.TON        = 1;         /* Timer on. */
   }
   
   if (1) /* Set up SPI2 and INT1 and INT2 on JOYMODE. */
   {
      /**
       * SPI2 responds to shift requests by the Amiga.
       * We set it up here, and then whenever the JOYMODE line is pulled
       * down, we enable SPI2 and put a valid state into the SPIxBUF for
       * the CD32 to read.
       * The CD32 supplies the clock on FIRE1OUT/CLOCKIN, Pin 22, RP11, RB11.
       * The data is sent out        on FIRE2OUT/DATAOUT, Pin 15, RP5,  RB5.
       * Only two pins are used for this shift transfer.
       */
      
      /* SPI setup for CD32 communications on right side. */      
      /* Disable SPI module before configuring. */
      SPI2STATbits.SPIEN   = 0;  
            
      /* Bit of OSCCON register controlling peripheral remapping lock. */
      const unsigned char OSCCON_LOCK = 0b01000000;
      
      /* Unlock the peripheral remapping. */
      __builtin_write_OSCCONL(OSCCON & ~(OSCCON_LOCK));
      
      /* Assign the remappable peripheral pins OUTPUT registers according to my layout: */
      _RP5R    = 10; /* SPI2 DATA  output to RP5 pin. */
      _RP11R   =  0; /* SPI2 CLOCK RP11 pin is used for SPI2 CLOCK input, don't direct any output there. */
      /* Assign the remappable peripheral pins INPUT, RP11 is input for SPI2 slave clock. */
      _SCK2R   = 11; /* SPI2 CLOCK input is directed to read from RP11. */
      /* Assign the remappable peripheral pins INPUT, RB10 is input for SPI2 SSxEN */
      _SS2R    = 10; /* SPI2 SSen  input is directed to read from RP10. */
      /* Assign the remappable peripheral pins INPUT, RB10 is input for INT1 and INT2. */
      _INT1R   = 10;
      _INT2R   = 10;

      /* Lock the peripheral remapping. */
      __builtin_write_OSCCONL(OSCCON | OSCCON_LOCK);

      /* Disable interrupts for safety, and clear interrupt flag. */
      IFS2bits.SPI2IF      = 0;
      IEC2bits.SPI2IE      = 0;
      
      /* These are zero by default anyway... */
      SPI2STATbits.SPISIDL = 0;     /* Continue in idle mode. */
      SPI2STATbits.SPIROV  = 0;     /* Clear overflow flag.   */
            
      SPI2CON1bits.DISSCK  = 0;     /* Slave mode. Set to zero. */
      SPI2CON1bits.DISSDO  = 0;     /* Enable data output. */      
      SPI2CON1bits.MODE16  = 1;     /* Exchange 16-bit words. */
      SPI2CON1bits.SMP     = 0;     /* Cleared since we are in slave mode. Sample in middle of clock cycle. */
      SPI2CON1bits.CKE     = 1;     /* Output data is changed on a rising clock edge. */
      SPI2CON1bits.SSEN    = 1;     /* SSx pin is used. */
      SPI2CON1bits.CKP     = 1;     /* Output data is changed on a rising clock edge. */
      SPI2CON1bits.MSTEN   = 0;     /* PIC24 chip is SPI Slave. */
      SPI2CON2bits.FRMEN   = 0;     /* Framed mode disabled. It isn't used in this protocol and not supported by the PIC24 anyway. */
      
      /* No clock rate to set. */
      /* Don't enable the module. */      
      
      /* Set up two interrupts on pin 21, CN16, RP10, RB10. */
      _TRISB10             = 1;     /* CD32 Joymode input. */
      INTCON2bits.INT1EP   = 1;     /* INT1 fires on the falling edge of RB10. */
      INTCON2bits.INT2EP   = 0;     /* INT2 fires on the  rising edge of RB10. */
      IPC5bits.INT1IP      = 5;     /* Give INT1 the highest priority of the interrupts I'm using. */
      IPC7bits.INT2IP      = 5;     /* Give INT2 the highest priority of the interrupts I'm using. */
      IFS1bits.INT1IF      = 0;     /* Clear flags. */
      IFS1bits.INT2IF      = 0;
      IEC1bits.INT1IE      = 1;     /* Enable interrupts. */
      IEC1bits.INT2IE      = 1;
      
      IFS2bits.SPI2IF      = 0;     /* Clear interrupt flag for SPI event. */
      IEC2bits.SPI2IE      = 1;     /* Enable interrupt request from SPI. */
   }
   
   global_pscd32_state.output_mode = PSCD32_OUTPUT_MODE_UNDEFINED;
   
   PSCD32_set_up_output_for_CD32_joypad_emulation();
    
   while (1)
   {            
      /* TODO: While we're not responding to an event, we can enter DOZE? */
      /* This means that every time the interrupt handling stack is empty, we always return to low-power? */
   }
   
   return 0;
}
